from circle.resources.abstract.createable import CreateableAPIResource
from circle.resources.abstract.listable import ListableAPIResource
from circle.resources.abstract.retrievable import RetrievableAPIResource
from circle.resources.abstract.updateable import UpdateableAPIResource



from circle import util
from circle.api_requestor import APIRequestor
from circle.resources.abstract.api_resource import APIResource


class ListableAPIResource(APIResource):
    @classmethod
    def list(cls, api_key=None, **params):
        requestor = APIRequestor(api_key)
        url = cls.class_url()
        response, api_key = requestor.request("get", url, params)
        circle_object = util.convert_to_circle_object(response, api_key)
        return circle_object



from circle import api_requestor, util
from circle.resources.abstract.api_resource import APIResource


class CreateableAPIResource(APIResource):
    @classmethod
    def create(cls, api_key=None, idempotency_key=None, **params):

        requestor = api_requestor.APIRequestor(api_key)
        url = cls.class_url()
        response, api_key = requestor.request("post", url, params)
        return util.convert_to_circle_object(
            response, api_key, klass_name=cls.OBJECT_NAME
        )



from circle.resources.abstract import (
    CreateableAPIResource,
    ListableAPIResource,
    RetrievableAPIResource
)


class Transfer(CreateableAPIResource, ListableAPIResource):
    """
    https://developers.circle.com/reference#payouts-transfers-create
    """

    OBJECT_NAME = "transfers"  # The object name as it maps to the API resource.



import json
import urllib
import uuid

import circle
from circle import http_client
from circle.circle_response import CircleResponse


class APIRequestor:
    def __init__(self, api_key=None, api_base=None, client=None):
        self.api_base = api_base or circle.api_base
        self.api_key = api_key or circle.api_key
        if client:
            self._client = client
        elif circle.default_http_client:
            """Users can change the default http client with circle.default_http_client"""
            self._client = circle.default_http_client
        else:
            """If no default http client is set, set one to avoid creating one for every request."""
            circle.default_http_client = http_client.new_default_http_client()
            self._client = circle.default_http_client

    def interpret_response(self, rbody, rcode, rheaders):
        try:

            resp = CircleResponse(rbody, rcode, rheaders)
        except Exception:
            raise Exception(
                f"Invalid response from API: {rbody} (HTTP response code: {rcode})",
                rbody,
                rcode,
                rheaders,
            )
        # if not 200 <= rcode < 300:
        #     self.handle_error_response(rbody, rcode, resp.data, rheaders)
        return resp

    def request_headers(self, api_key):
        user_agent = "Circle Python SDK"
        headers = {
            "Authorization": "Bearer %s" % api_key,
            "User-Agent": user_agent,
            "Content-Type": "application/json",
            "Accept": "application/json",
        }
        return headers

    def request(self, method, url, params, headers=None):
        rbody, rcode, rheaders, api_key = self.request_raw(method.lower(), url, params)
        resp = self.interpret_response(rbody, rcode, rheaders)
        return resp, api_key

    def request_raw(self, method, url, params={}, supplied_headers=None):
        if self.api_key:
            api_key = self.api_key
        else:
            from circle import api_key

            api_key = api_key
        if api_key is None:

            # TODO: use custom AuthenticationError exception here
            raise Exception("Invalid api key")

        abs_url = "%s%s" % (self.api_base, url)

        if method in (
            "post",
            "patch",
        ):
            params.setdefault("idempotencyKey", uuid.uuid4().__str__())
            post_data = json.dumps(params)
        elif method in (
            "get",
            "delete",
        ):
            post_data = None
            # TODO: generate query params for get/delete methods
        else:
            # TODO: use a proper error here
            raise Exception("Invalid HTTP method.")

        headers = self.request_headers(api_key)
        if supplied_headers is not None:
            for key, val in supplied_headers.items():
                headers[key] = val

        rbody, rcode, rheaders = self._client.request_with_retries(
            method, abs_url, headers, post_data
        )

        return rbody, rcode, rheaders, api_key



import circle
from circle.circle_response import CircleResponse


def get_object_classes():
    from circle.object_classes import OBJECT_CLASSES

    return OBJECT_CLASSES


def convert_to_circle_object(response, api_key=None, klass_name=None):

    # If a CircleResponse is passed, return the StripeObject associated with it.

    if isinstance(response, CircleResponse):
        circle_response = response
        response = circle_response.data
    if isinstance(response, list):
        return [convert_to_circle_object(r, api_key) for r in response]
    elif get_object_classes().get(klass_name) is not None:
        return get_object_classes().get(klass_name).construct_from(response, api_key)
    elif isinstance(response, dict) and not isinstance(
        response, circle.circle_object.CircleObject
    ):

        return circle.circle_object.CircleObject.construct_from(response, api_key)
    elif isinstance(response, dict):
        return circle.circle_object.CircleObject.construct_from(response, api_key)
    else:
        return response



from circle import resources

OBJECT_CLASSES = {
    resources.Ach.OBJECT_NAME: resources.Ach,
    resources.Card.OBJECT_NAME: resources.Card,
    resources.Payment.OBJECT_NAME: resources.Payment,
    resources.Payout.OBJECT_NAME: resources.Payout,
    resources.Wire.OBJECT_NAME: resources.Wire,
}



import json


class CircleResponse:
    def __init__(self, body, code, headers):
        self.body = body
        self.code = code
        self.headers = headers

        json_body = json.loads(body)
        self.data = json_body.get("data") or json_body

    @property
    def idempotency_key(self):
        try:
            return self.headers["idempotency-key"]
        except KeyError:
            return None



import threading
import time

import requests


def new_default_http_client(*args, **kwargs):
    if requests:
        return RequestsClient(*args, **kwargs)
    else:
        raise Exception(
            "Requests must be installed as it is currently the only option for an HTTP client."
        )


class HttpClient:
    MAX_DELAY = 2  # in seconds
    INITIAL_DELAY = 0.5  # in seconds

    def __init__(self):
        self._thread_local = threading.local()

    def request_with_retries(self, method, url, headers, post_data=None):
        num_retries = 0
        while True:
            try:
                response = self.request(method, url, headers, post_data)
                connection_error = None
            except Exception as e:  # TODO: use custom error
                response = None
                connection_error = e

            if self._should_retry(response, connection_error, num_retries):
                num_retries += 1
                sleep_time = self._sleep_time_seconds(num_retries, response)
                time.sleep(sleep_time)
            else:
                if response is not None:
                    return response
                else:
                    raise Exception  # TODO: raise connection error

    def request(self, method, url):
        raise NotImplementedError(
            f"HttpClient sublcass must implement `request` method."
        )

    @property
    def _max_network_retries(self):
        from circle import max_network_retries

        return max_network_retries

    def _should_retry(self, response, api_connection_error, num_retries):
        if num_retries > self._max_network_retries:
            return False

        if response is None:
            # TODO: we eventually want the subclasses to handle this. for now, default to not retry on connection
            # issues/timeouts.
            return False

        content, status_code, rheaders = response

        if status_code >= 500:
            print("should retry...")
            return True
        return False

    def _sleep_time_seconds(self, num_retries, response=None):
        sleep_seconds = min(
            self.MAX_DELAY, self.INITIAL_DELAY * (2 ** (num_retries - 1))
        )  # Double delay with each retry until we reach the max delay

        return sleep_seconds


class RequestsClient(HttpClient):
    name = "requests"

    def __init__(self, timeout=30, **kwargs):
        self._timeout = timeout
        self._session = None
        super().__init__(**kwargs)

    def request(self, method, url, headers, post_data=None):

        if getattr(self._thread_local, "session", None) is None:
            self._thread_local.session = self._session or requests.Session()

        try:
            res = self._thread_local.session.request(
                method,
                url,
                headers=headers,
                data=post_data,
                timeout=self._timeout,
            )
        except Exception as e:  # TODO: update exception
            raise e

        return res.content, res.status_code, res.headers



api_key = None
api_environment = None
default_http_client = None
api_base = {
    "sandbox": "https://api-sandbox.circle.com",
    "production": "https://api.circle.com",
}.get(api_environment if api_environment else "sandbox")


max_network_retries = 0

from circle.resources import *



from circle.resources.abstract.api_resource import APIResource
from circle.api_requestor import APIRequestor
from circle import util


class PublicKey(APIResource):
    """
    The public key is obtained with the .retrieve() method which is supported in the base APIResource
    https://developers.circle.com/reference/encryption
    """

    OBJECT_NAME = "encryption.public"  # The object name as it maps to the API resource.

    @classmethod
    def retrieve(cls, api_key=None, **kwargs):
        """
        The public key retrieve method looks like the list method as there is no ID to query by.

        However, this returns a {} not a [], thus, for consistency we want devs to call the retrieve method.
        """
        requestor = APIRequestor(api_key)
        url = cls.class_url()
        response, api_key = requestor.request("get", url, kwargs)
        circle_object = util.convert_to_circle_object(response, api_key)
        return circle_object



import circle


class CircleAPIError(Exception):
    """
    https://developers.circle.com/docs/api-errors
    """

    def __init__(
        self, message=None, body=None, status_code=None, headers=None, code=None
    ):
        self.message = message
        self.body = body
        self.status_code = status_code
        self.headers = headers or {}
        self.code = code
        self.request_id = self.headers.get("X-Request-Id", None)

    def __str__(self):
        msg = self.message
        if self.request_id:
            return f"Request {self.request_id}: {msg}"
        else:
            return msg

    def __repr__(self):
        return f"{self.__class__.__name__}(message={self.message}, status_code={self.status_code}, request_id={self.request_id})"


class AuthenticationError(CircleAPIError):
    pass


# Entity Errors:

# TODO: this might not actually be suited as an exception. Will come back to it.
class CircleEntityErrors(Exception):

    """
    https://developers.circle.com/docs/api-errors
    """

    pass



from circle import api_requestor, util


class CircleObject(dict):
    def __init__(self, id=None, api_key=None, **params):
        super().__init__()

        self._unsaved_values = (
            set()
        )  # Values that haven't change in a refresh and should persist to the new state
        self._transient_values = set()
        self._retrieve_params = params
        # TODO: I forget why I did this?
        for k, v in params.items():
            self.__setattr__(k, v)

        self.__setattr__("api_key", api_key)

        if id:
            self["id"] = id

    def __setattr__(self, k, v):
        if k[0] == "_" or k in self.__dict__:
            return super().__setattr__(k, v)
        self[k] = v
        return None

    def __getattr__(self, k):
        try:
            return self[k]
        except KeyError as e:
            raise AttributeError(e)

    @classmethod
    def construct_from(cls, values, key):
        instance = cls(values.get("id"), api_key=key)
        instance.refresh_from(values, key)
        return instance

    def request(self, method, url, params=None, headers=None):

        requestor = api_requestor.APIRequestor(api_key=self.api_key)
        response, api_key = requestor.request(method, url, params, headers)

        return util.convert_to_circle_object(response, api_key)

    def refresh_from(self, values, api_key=None, partial=False, last_response=None):
        self.api_key = api_key or getattr(values, "api_key", None)
        if partial:
            # Record values that aren't set but, since this is partial=True, will persist to the new state
            self._unsaved_values = self._unsaved_values - set(values)
        else:
            removed = set(self.keys()) - set(values)
            self._transient_values = (
                self._transient_values | removed
            )  # include existing transient values and any removed values
            self._unsaved_values = (
                set()
            )  # This is not a partial update, so all state will be refreshed.
            self.clear()

        self._transient_values = self._transient_values - set(values)

        for k, v in values.items():
            super().__setitem__(
                k,
                util.convert_to_circle_object(
                    v, api_key, klass_name=self.__class__.__name__
                ),
            )



from circle.resources.ach import Ach
from circle.resources.card import Card
from circle.resources.payment import Payment
from circle.resources.payout import Payout
from circle.resources.public_key import PublicKey
from circle.resources.transfer import Transfer
from circle.resources.wire import Wire
